//------------------------------------------------------------------------
//
// OpenDCC - OpenDecoder2
//
// Copyright (c) 2006,2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      dimm_engine.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-02-20 V0.1 kw copied from opendecoder.c
//                               removed code for versatile output -
//                               we only do turnout commands and permanent
//                               commands (up to now)
//            2007-04-10 V0.2 kw added init code for timer 1
//                               turn off output only if time > 0
//                               outputs with time = 0 are kept forever
//            2007-04-23 V0.3 kw added compile switch
//                               PORT_ENABLED
//            2007-04-27 V0.4 kw evaluation of ReceivedActivate
//            2007-05-14 V0.5 kw added masking of TOIE1 to prevent reentry
//            2007-05-25 V0.6 kw Bugfix in init of FEEDBACK_PULLUP
//            2007-10-29 V0.7 kw added belgian Signals
//            2007-05-30 V0.8 kw added reinit of states (except blinking)
//
// tests:     2007-04-14 kw: feedback tested, FBM = 0,1; Magnet coils
//
//------------------------------------------------------------------------
//
// purpose:   flexible general purpose decoder for dcc
//            here: timing engine for outputs
//                  software pwm and state engine
//
// content:   A DCC-Decoder for ATmega8515 and other AVR
//
//             1. Defines and variable definitions
//             2. Init
//             3. ISR
//             4. Predefined values
//             5. action - performing the Command from the DCC call
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "hardware.h"
#include "dcc_receiver.h"
#include "main.h"
#include "dimm_engine.h"

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine
                                 // 3: test action

#define DIMM_ENABLED  TRUE

//=============================================================================
// 1. Definitions
//=============================================================================
//
// Output:
// Port Engine does the control of the outputs only if PORT_ENABLED is true
//
#ifndef PORT_ENABLED
  #warning PORT_ENABLED not defined, now enabled by default (should be done in config.h)
  define PORT_ENABLED  TRUE
#endif

// Definitions for the DIMM-Engine
// see below on information about the DIMM-Engine

#define PWM_TICK_PERIOD     300L        // 300us tick for PWM Engine
#define PWM_STEPS           60L
#define PWM_PERIOD          (PWM_TICK_PERIOD * PWM_STEPS)    // 18ms = 300 * 60L
#define DIMM_UP_DELAY       400000L     // Dimm up with 400ms delay -> now in eeprom
#define DIMM_DOWN_SPEED     4L          // Dimm down slow 
#define DIMM_UP_SPEED       5L          // Dimm up a little faster -> now in eeprom
#define DIMM_DOWN_SAME_BULB 1           // 1: wenn im neuen Begriff die lampe 
                                        //    wieder leuchtet, zuerst abdimmen,
                                        //    dann wieder aufdimmen.
                                        // 0: leuchtet durchgehend  

#define DIMM_RANGE_MIN 100    // aktiver Bereich 100-160
#define DIMM_RANGE_MAX (DIMM_RANGE_MIN+PWM_STEPS+1)


//----------------------------------------------------------------------------
// Global Data
// out_pwm controls outputs, monoflop pulses and toggling operations
//
//------------------------------------------- PWM

// a) running values

unsigned char cur_dimm;            //  = DIMM_RANGE_MIN;

unsigned char CurrentTarget;


// b) target values
volatile struct
  {
    unsigned char rest;     	// Zeit bis zum nchsten Wechsel in PWM_PERIOD (18ms)
    unsigned char ontime;   	// Einschaltzeit
    unsigned char offtime;      // Ausschaltzeit
    unsigned char dimm_val;     // aktueller Istwert
    unsigned char light_A_val;  // aktueller Zielwert in Richtung A
    unsigned char delta_A;      // aktuelles Increment in Richtung A
    unsigned char light_B_val;  // aktueller Zielwert in Richtung B
    unsigned char delta_B;      // aktuelles Decrement in Richtung B
  } out_pwm[8];
#define SIZE_out_pwm (sizeof(out_pwm) / sizeof(out_pwm[0]))

unsigned char DimmUpDelay;          // given in Ticks (and loaded from eeprom);
unsigned char BlinkPeriod_2;        // if a blink bit is set, use this value to determine period

// control structure for turnouts
struct
  {
    unsigned char pulse_duration;
    unsigned char position;         // red (=0) or green (=1)
    enum fb_modes                   // mirror of cv546 ... cv549
      {                           
         COMMAND_ONLY = 0,          // 0:  no feedback, just a mirror of the command
         END_SWITCH = 1,            // 1:  feedback through intrinsic 
                                    //     coil end switches (like with roco line)
         EXTERN_LOW = 2,            // 2:  feedback with extra indicator -
                                    //     PC0 is low, when in *red* position
         EXTERN_HIGH = 3            // 3:  feedback with extra indicator -
                                    //     PC0 is high, when in *red* position
      } feedback_mode;        
  } turnout[4];


unsigned char failed_turnout;       // this is the last turnout with an error [0..3]

enum feedback_methodes              // a mirror of cv545
  {
    NO_FB = 0,
    POS_ACK = 1,
    ALARM = 2,
  } feedback_method;

//----------------------------
volatile struct
  {
    unsigned char rest;     // Zeit bis zum wechsel in Ticks (20ms)
    unsigned char ontime;   // Einschaltzeit
    unsigned char offtime;  // Ausschaltzeit
    unsigned char pause;    
    unsigned char flashes;  // Anzahl Pulse
    unsigned char act_flash;    
  } led;


void turn_led_on(void)
  {
    led.rest = 0;
    LED_ON;
  }

void turn_led_off(void)
  {
    led.rest = 0;
    LED_OFF;
  }


// make a series of flashes, then a longer pause
void flash_led_fast(unsigned char count)
  {
    led.act_flash = 1;
    led.flashes = count;
    led.pause   = 700000L / TICK_PERIOD;
    led.offtime = 240000L / TICK_PERIOD;
    led.ontime  = 120000L / TICK_PERIOD;
    led.rest    = 120000L / TICK_PERIOD;
    LED_ON;
  }


// some static used variables

unsigned char hauptsignal_state;    // fr Dunkeltastung
unsigned char vorsignal_state;


// Trick:   Speed up port access by declaring port access as inline code
//          This is done with a subroutine; if gcc runs with -os, this results in
//          single cbi and sbi statements!
// Note:    OUTPUT_PORT is defined in hardware.h


void output(unsigned char port_no, unsigned char state)
  {
    if (state == 0)
      {                              
        OUTPUT_PORT &= (~(1<<port_no));
      }
    else
      {
        OUTPUT_PORT |= (1<<port_no);
      }
  }
                 
static inline void disable_timer_interrupt(void)   __attribute__((always_inline));
void disable_timer_interrupt(void)
  {
      // TIMSK &= ~(1<<TOIE1);        // Timer1 Overflow
      TIMSK &= ~(1<<OCIE1A);        // Timer1 Compare A
  }

static inline void enable_timer_interrupt(void)   __attribute__((always_inline));
void enable_timer_interrupt(void) 
  {
      // TIMSK |= (1<<TOIE1);        // Timer1 Overflow
      TIMSK |= (1<<OCIE1A);        // Timer1 Compare A
  }


static inline void set_target(unsigned char port, unsigned char state) 
       __attribute__((always_inline));

void set_target(unsigned char port, unsigned char state)
  {
    if (state == 0)
      {                              
        CurrentTarget &= (~(1<<port));
      }
    else
      {
        CurrentTarget |= (1<<port);
      }
  }

//==============================================================================
//
// Section 2
//
//------------------------------------------------------------------------------
//
// init_dimm_engine(void)
//   initializes the output of the decoder 
//   a) setup timer (with TICK_PERIOD)
//   b) load values for port states from eeprom
//   c) preload outputs

void init_dimm_engine(void)
  {

    unsigned char i, j;


    #if (DIMM_ENABLED == TRUE)
    // Init Timer1 as CTC 
    // check PWM_TICK_PERIOD and F_CPU
    #define TIMER1_CLKDIV 64              // possible values: 1, 8, 64, 256, 1024

    #if (F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER1_CLKDIV) > 255L
      #warning: overflow in OCR1A - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a larger clkdiv
    #endif    
    #if (F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER1_CLKDIV) < 30L
      #warning: resolution accuracy in OCR1A too low - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a smaller clkdiv
    #endif    


    OCR1A = F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER1_CLKDIV ;  
    TCCR1A = (0 << COM1A1)      // compare match A (disconnected)
           | (0 << COM1A0) 
           | (0 << COM1B1)      // compare match B (disconnected)
           | (0 << COM1B0) 
           | (0 << FOC1A)                  // reserved
           | (0 << FOC1B)                  // reserved
           | (0 << WGM11)  
           | (0 << WGM10);      // Timer0 Mode 4 = CTC - Int on compare A (Top = OCR1A)
    TCCR1B = (0 << ICNC1) 
           | (0 << ICES1) 
           | 0                // reserved
           | (0 << WGM13) 
           | (1 << WGM12) 
    #if   TIMER1_CLKDIV == 1
           | (0 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[2:0]=001 clkdiv 1
    #elif TIMER1_CLKDIV == 8
           | (0 << CS12) | (1 << CS11)  | (0 << CS10);       // CS[2:0]=010 clkdiv 8
    #elif TIMER1_CLKDIV == 64
           | (0 << CS12) | (1 << CS11)  | (1 << CS10);       // CS[2:0]=011 clkdiv 64 
    #elif TIMER1_CLKDIV == 256
           | (1 << CS12) | (0 << CS11)  | (0 << CS10);       // CS[2:0]=100 clkdiv 256
    #elif TIMER1_CLKDIV == 1024
           | (1 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[2:0]=101 clkdiv 1024   
    #else
     #warning: TIMER1_CLKDIV is void
    #endif

    TIMSK |= (0<<TOIE1)             // Timer1 Overflow (-> was 1)
           | (1<<OCIE1A)            // Timer1 Compare A
           | (0<<OCIE1B)            // Timer1 Compare B
           | 0                      // reserved
           | (0<<TICIE1)            // Timer1 Input Capture
           | 0                      // reserved (Timer0 Compare B)
           | (0<<TOIE0)             // Timer0 Overflow
           | (0<<OCIE0);            // Timer0 Compare A
      

    timerval = 0;

    // now init all state variables
      
    cur_dimm = DIMM_RANGE_MIN;
        
    #endif


//   b) load values for port states from eeprom


// Delta fr Glhlampensimulation vorbelegen

    j = eeprom_read_byte(&CV.DimmUpSpeed);
    if (j==0) j=1;    if (j>PWM_STEPS+2) j=PWM_STEPS+2;
    for (i=0; i<SIZE_out_pwm; i++)
      {
        out_pwm[i].delta_A     = j;
      }

    j = eeprom_read_byte(&CV.DimmDownSpeed);
    if (j==0) j=1;    if (j>PWM_STEPS+2) j=PWM_STEPS+2;
    for (i=0; i<SIZE_out_pwm; i++)
      {
        out_pwm[i].delta_B      = j;
      }
          
    j = eeprom_read_byte(&CV.TargetUp) + DIMM_RANGE_MIN;            // add Offset for engine
    for (i=0; i<SIZE_out_pwm; i++)
      {
        out_pwm[i].light_A_val    = j;
      }

    j = eeprom_read_byte(&CV.TargetDown) + DIMM_RANGE_MIN;            // add Offset for engine
    for (i=0; i<SIZE_out_pwm; i++)
      {
        out_pwm[i].light_B_val   = j;
        out_pwm[i].dimm_val   = j;
      }

    DimmUpDelay = eeprom_read_byte(&CV.DimmUpDelay);
    BlinkPeriod_2 = eeprom_read_byte(&CV.BlinkPeriod_2);


    #if (PORT_ENABLED == TRUE)
      {
        unsigned char last_state, mask, i;
        unsigned char make_pullup = 0;
  
        // get cv's from eeprom

        turnout[0].pulse_duration = eeprom_read_byte(&CV.T_on_F1);
        turnout[0].feedback_mode =  eeprom_read_byte(&CV.FBM_F1);   // cv546
        turnout[1].pulse_duration = eeprom_read_byte(&CV.T_on_F2);
        turnout[1].feedback_mode =  eeprom_read_byte(&CV.FBM_F2);   // cv547
        turnout[2].pulse_duration = eeprom_read_byte(&CV.T_on_F3);
        turnout[2].feedback_mode =  eeprom_read_byte(&CV.FBM_F3);   // cv548
        turnout[3].pulse_duration = eeprom_read_byte(&CV.T_on_F4);
        turnout[3].feedback_mode =  eeprom_read_byte(&CV.FBM_F4);   // cv549

        feedback_method = eeprom_read_byte(&CV.FM);    // cv545
        failed_turnout = 0x80;   // nothing wrong


        // check out for last states, but only if we do direct port operations

        if (eeprom_read_byte(&CV.MODE) == 0)    // direct port mode
          {
            last_state = eeprom_read_byte(&CV.LastState);
            mask = 0b000011;

            for (i=0; i<4; i++)
              {
                if ((turnout[i].feedback_mode == EXTERN_LOW) || (turnout[i].feedback_mode == EXTERN_HIGH))
                  {
                    // this needs pullup
                    make_pullup |= mask;
                  }
  
                if (turnout[i].pulse_duration != 0)
                  {
                    // this is pulsed, no last state
                    last_state &= ~mask;
                  }
                mask = mask << 2;
              }

            FEEDBACK_PULLUP = make_pullup;              // this is write only
            OUTPUT_PORT = last_state;
            PortState = last_state;  
          }        
      }
    #endif


//   c) preload outputs and states

    CurrentTarget = eeprom_read_byte(&CV.LastState);
    hauptsignal_state = eeprom_read_byte(&CV.HauptState);
    vorsignal_state = eeprom_read_byte(&CV.VorState);

    switch (eeprom_read_byte(&CV.LightMode))
      {
        default:
        case  0: // 00 = 4 pairs, switched, 8 addresses (traditional)
                 //      (duration according to T_on_xx)
                for (i=0; i<SIZE_out_pwm; i++)
                  {
                    out_pwm[i].delta_A     = PWM_STEPS+2;
                    out_pwm[i].delta_B     = PWM_STEPS+2;
                  }
                break;
        case  1: // 01 = 8 addresses, dimmed, static states
                break;
        case  2: // 02 = 16 addresses, dimmed, static states
                break;
                
        case  8: // 08 = traffic lights (Ampel) 8 addresses
                out_pwm[3].delta_A = PWM_STEPS+2;   // Stopmagnet soll durchschalten
                out_pwm[3].delta_B = PWM_STEPS+2; 
 	            out_pwm[7].delta_A = PWM_STEPS+2;   // Stopmagnet soll durchschalten
                out_pwm[7].delta_B = PWM_STEPS+2;
                action(hauptsignal_state, 1);          // restore old state at power up
                action(vorsignal_state, 1);
                break;
        case  9: // 09 = rail crossing (Andreaskreuz) 8 addresses
                out_pwm[3].delta_A = PWM_STEPS+2; // Stopmagnet soll durchschalten
                out_pwm[3].delta_B = PWM_STEPS+2; 
 	            out_pwm[7].delta_A = PWM_STEPS+2; // Stopmagnet soll durchschalten
                out_pwm[7].delta_B = PWM_STEPS+2;
                action(hauptsignal_state, 1);          // restore old state at power up
                action(vorsignal_state, 1);                
                break;
        case 10: // 10 = Flasher
                action(hauptsignal_state, 1);          // restore old state at power up
                break;
        case 16: // 16 = Lichtsule (aus, langsam, schnell)
                action(hauptsignal_state, 1);          // restore old state at power up
                break;
        case 17: // 17 = Lichterkette vorwrts (aus, eine langsam, zwei langsam, drei,  schnell)
                action(hauptsignal_state, 1);          // restore old state at power up
                break;
        case 18: // 18 = Lichterkette rckwrts (aus, eine langsam, zwei langsam, drei,  schnell)
                action(hauptsignal_state, 1);          // restore old state at power up
                break;
        case 19: // 19 = Baustellenblitzer
                action(hauptsignal_state, 1);          // restore old state at power up
                break;
      } 
  }

//==============================================================================
//
// Section 3
//
// Timing Engine
//
// Howto:    Timer1 (with prescaler 8 and 16 bit total count) triggers
//           an interrupt every PWM_PERIOD (=300us @8MHz);
// 
//==============================================================================
//
// Section 3
//
// Lichtsteuerung als PWM-Dimmer
//
//             |                        |                          |
// Output:     |XXXXXXXXXX______________|XXXXXXXXXX________________|
//             |<-------->              |<------------------------>|
//             | dimm_val               |        PWM_PERIOD        |
//         ----|------------------------|--------------------------|---> Time
//
//
// 1. Einstellen des aktuellen Helligkeitswertes (do_dimm)
//
//    Es gibt 60 Helligkeitstufen.
//    Alle 300us erfolgt ein Interrupt, dieser schaltet den Dimmer um eine
//    Stufe weiter, nach 60 Stufen wird wieder von vorne begonnen.
//    Stufe MIN:  alle Ports mit einem dimm_val > DIMM_MIN werden eingeschaltet.
//    Stufe x:    Ein Port, dessen dimm_val kleiner x ist, wird abgeschaltet.
//    Stufe MAX:  Restart und Meldung an den DIMMER (C_Dimmstep)
//
//    Folge: Alle Ports mit einem dimm_val kleiner DIMM_RANGE_MIN sind
//    dauerhaft aus, alle Ports mit einem dimm_val grer DIMM_RANGE_MAX
//    sind dauerhaft ein.
//
//    Da der PWM 60 Stufen hat und alle 300us ein Int erfolgt, wird dieser
//    Durchlauf alle 18ms durchgefhrt. Dies entspricht einer Refreshrate von
//    55Hz.
//
//
// 2. Langsame Vernderung der Helligkeit
//
//    Nach einem Zyklus des PWM wird vom Hauptprogramm der neue aktuelle
//    dimm_val ausgerechnet. Hierzu wird vom aktuellen Wert mit einem
//    Schritt "delta" nach oben oder unten gerechnet, bis der neue Zielwert
//    erreicht ist.
//
//    Die Zykluszeit der PWM ist 18ms, somit wird je nach "delta" folgende
//    Dimmzeit erreicht:
//
//      delta    |    Dimmzeit
//    -----------------------------
//        1      |    1080ms
//        2      |     540ms
//        3      |     360ms
//        4      |     270ms
//        5      |     216ms
//        6      |     180ms
//      100      |    sofort
//
//    Der neue Zielwert wird in light_val hinterlegt. 
//
//    Wenn man von einem Wert kleiner DIMM_RANGE_MIN startet, dann wird
//    der Port erst mit Verzgerung aufgedimmt, weil zuerst der Bereich
//    bis DIMM_RANGE_MIN "aufgedimmt" wird.
//    Dies wird dazu benutzt, zuerst das alte Signalbild wegzudimmen 
//    und dann das neue Signalbild aufzudimmen.
//
// 3. Signalbilder
// 
//    Signalbilder werden als Bitfeld hinterlegt. Fr jedes Kommando
//    gibt es ein Bitfeld, in dem das neue Signalbild abgelegt ist und eine
//    Gltigkeitsmaske, diese bestimmt, auf welche Dimmwerte das Signalbild
//    wirken soll. Mit diesen beiden Pattern wird "set_new_light_val" 
//    aufgerufen.
//
// 4. Blinken
//    
//    Falls ontime bzw. offtime ungleich 0 sind, wird nach Ablauf der
//    jeweils andere Phasenwert geladen.
//



// This is the PWM Interrupt

ISR(TIMER1_COMPA_vect)                      // Timer1 Compare Int
  {                                         // macht pwm
    unsigned char port;
    unsigned char mask;

    disable_timer_interrupt();
    sei();                                  // allow dcc receive
    mask = 1;
    cur_dimm++;
    if (cur_dimm == DIMM_RANGE_MAX)
      {
        cur_dimm = DIMM_RANGE_MIN;
        for (port=0; port<8; port++)
          {
            if (out_pwm[port].dimm_val > DIMM_RANGE_MIN) PORTB |= mask;   // Einschalten wenn !0
            mask = mask << 1; 
          }
        Communicate |= (1<<C_Tick);
        timerval++;                         // advance global tick
      }
    else
      {
        for (port=0; port<8; port++)
          {
            if (cur_dimm >= out_pwm[port].dimm_val) PORTB &= ~mask; 
            mask = mask << 1;
          }
      }
    enable_timer_interrupt();
  }




#define NORMAL_ISR FALSE

#if (NORMAL_ISR == TRUE)

ISR(TIMER1_OVF_vect)                        // Timer1 Overflow Int
  {
    
    disable_timer_interrupt();
    
    sei();                                  // allow DCC interrupt

    timerval++;                             // advance global clock

    #if (PORT_ENABLED == TRUE)
      {
        unsigned char port;
        unsigned char mask; 
        unsigned char my_rest; 
       
        mask = 1;
        for (port=0; port<8; port++)
          {
            my_rest = out_pwm[port].rest;       // use a local variable to force
            if (my_rest !=0)                    // compiler to tiny code
              {
                if (--my_rest == 0)
                  {
                     if (OUTPUT_PORT & mask)
                      { // bit was on
                        my_rest = out_pwm[port].offtime;
                        OUTPUT_PORT &= ~mask;
                      }
                    else
                      {
                        my_rest = out_pwm[port].ontime;
                        OUTPUT_PORT |= mask;
                      }
                  }
                out_pwm[port].rest = my_rest;
              }
             mask = mask << 1;                  // do *not* calc mask from port
          }
      }
    #endif

    // now process led

    if (led.rest != 0)
      {
        if (--led.rest == 0)
          {
             if (LED_STATE)
              {
                if (led.act_flash == led.flashes)
                  {
                    led.rest = led.pause;
                    led.act_flash = 0;
                  }
                else
                  {
                    led.rest = led.offtime;
                  }
                LED_OFF;
              }
            else
              {
                led.act_flash++;
                led.rest = led.ontime;
                LED_ON;
              }
          }
      }
    enable_timer_interrupt();
  }
#endif


//==============================================================================
//
// Section 4
//
// Timing Values for the Timing Engine
//
// All output control instructions are stored in tables. These tables are
// evaluated, when a action a pin is be performed.
//
// If a port is never toggled from the timing engine, it my be directly set.
//
struct t_action
  {
    unsigned char status;           // Multi purpose byte:
                                    // Bit 7:    new state of output pin
                                    // Bit 6:    variable first pulse enabled
                                    // Bit 2..0: output pin number
    unsigned char rest;             // time to keep the state; if 0: keep forever
    unsigned char ontime;           // ontime: if elapsed, output will turn off
    unsigned char offtime;          // 
  };


/*------------- Some predefined structures

struct t_action port0_on PROGMEM =   { (1 << 7) + 0, 0, 0, 0};
struct t_action port1_on PROGMEM =   { (1 << 7) + 1, 0, 0, 0};
struct t_action port2_on PROGMEM =   { (1 << 7) + 2, 0, 0, 0};
struct t_action port3_on PROGMEM =   { (1 << 7) + 3, 0, 0, 0};
struct t_action port4_on PROGMEM =   { (1 << 7) + 4, 0, 0, 0};
struct t_action port5_on PROGMEM =   { (1 << 7) + 5, 0, 0, 0};
struct t_action port6_on PROGMEM =   { (1 << 7) + 6, 0, 0, 0};
struct t_action port7_on PROGMEM =   { (1 << 7) + 7, 0, 0, 0};

struct t_action port0_off PROGMEM =  { (0 << 7) + 0, 0, 0, 0};
struct t_action port1_off PROGMEM =  { (0 << 7) + 1, 0, 0, 0};
struct t_action port2_off PROGMEM =  { (0 << 7) + 2, 0, 0, 0};
struct t_action port3_off PROGMEM =  { (0 << 7) + 3, 0, 0, 0};
struct t_action port4_off PROGMEM =  { (0 << 7) + 4, 0, 0, 0};
struct t_action port5_off PROGMEM =  { (0 << 7) + 5, 0, 0, 0};
struct t_action port6_off PROGMEM =  { (0 << 7) + 6, 0, 0, 0};
struct t_action port7_off PROGMEM =  { (0 << 7) + 7, 0, 0, 0};

//                                        "ON"     "PULS"  NR  
struct t_action port0_pulse PROGMEM =   { (1<<7) + (1<<6) + 0, 0, 0, 0};
struct t_action port1_pulse PROGMEM =   { (1<<7) + (1<<6) + 1, 0, 0, 0};
struct t_action port2_pulse PROGMEM =   { (1<<7) + (1<<6) + 2, 0, 0, 0};
struct t_action port3_pulse PROGMEM =   { (1<<7) + (1<<6) + 3, 0, 0, 0};
struct t_action port4_pulse PROGMEM =   { (1<<7) + (1<<6) + 4, 0, 0, 0};
struct t_action port5_pulse PROGMEM =   { (1<<7) + (1<<6) + 5, 0, 0, 0};
struct t_action port6_pulse PROGMEM =   { (1<<7) + (1<<6) + 6, 0, 0, 0};
struct t_action port7_pulse PROGMEM =   { (1<<7) + (1<<6) + 7, 0, 0, 0}; */



/*/                                           "ON" NR       rest,   ontime, offtime
struct t_action port0_amp1_rot    PROGMEM = { (0<<7) + 0,  AMP_YE, 0,      0   };
struct t_action port1_amp1_rot    PROGMEM = { (1<<7) + 1,  AMP_YE, 0,      0   };
struct t_action port3_amp1_rot    PROGMEM = { (0<<7) + 3,AMP_YE/2, 0,      0   };

struct t_action port0_amp1_green  PROGMEM = { (1<<7) + 0,  AMP_YE, 0,      0   };
struct t_action port1_amp1_green  PROGMEM = { (1<<7) + 1,  AMP_YE, 0,      0   };
struct t_action port2_amp1_green  PROGMEM = { (0<<7) + 2,  AMP_YE, 0,      0   };
struct t_action port3_amp1_green  PROGMEM = { (1<<7) + 3,3*AMP_YE/2, 0,      0   };

struct t_action port1_amp1_blink  PROGMEM = { (1<<7) + 1,  AMP_BL, AMP_BL, AMP_BL };

//
struct t_action port4_amp2_rot    PROGMEM = { (0<<7) + 4,  AMP_YE, 0,      0   };
struct t_action port5_amp2_rot    PROGMEM = { (1<<7) + 5,  AMP_YE, 0,      0   };
struct t_action port7_amp2_rot    PROGMEM = { (0<<7) + 7,AMP_YE/2, 0,      0   };

struct t_action port4_amp2_green  PROGMEM = { (1<<7) + 4,  AMP_YE, 0,      0   };
struct t_action port5_amp2_green  PROGMEM = { (1<<7) + 5,  AMP_YE, 0,      0   };
struct t_action port6_amp2_green  PROGMEM = { (0<<7) + 6,  AMP_YE, 0,      0   };
struct t_action port7_amp2_green  PROGMEM = { (1<<7) + 7,3*AMP_YE/2, 0,      0   };

struct t_action port5_amp2_blink  PROGMEM = { (1<<7) + 5,  AMP_BL, AMP_BL, AMP_BL };
*/


// Nachfolgend mal ein Baustellenblitzer
// LED0: ____x______________________________________x_____________________________
// LED1: _______x______________________________________x__________________________
// LED2: __________x______________________________________x_______________________
// LED3: _____________x______________________________________x____________________
// LED4: ________________x______________________________________x_________________
// LED5: ___________________x______________________________________x______________
// LED6: ______________________x______________________________________x___________
// LED7: _________________________x______________________________________x________



// #define BST_ON 1     // Time for one flash
// #define BST_OFF 80   // BST_ON + BST_OFF = period
// #define BST_RUN 4    // Offset between succeeding flashes
#define BST_ON  (  30000L / TICK_PERIOD)     // 30ms: Time for one flash
#define BST_OFF (1100000L / TICK_PERIOD)     // BST_ON + BST_OFF = period
#define BST_RUN ( 100000L / TICK_PERIOD)     // Offset between succeeding flashes

/*/                                        "ON" NR       rest, ontime, offtime
struct t_action port0_bst PROGMEM =   { (0<<7) + 0,   BST_RUN, BST_ON, BST_OFF};
struct t_action port1_bst PROGMEM =   { (0<<7) + 1, BST_RUN*2, BST_ON, BST_OFF};
struct t_action port2_bst PROGMEM =   { (0<<7) + 2, BST_RUN*3, BST_ON, BST_OFF};
struct t_action port3_bst PROGMEM =   { (0<<7) + 3, BST_RUN*4, BST_ON, BST_OFF};
struct t_action port4_bst PROGMEM =   { (0<<7) + 4, BST_RUN*5, BST_ON, BST_OFF};
struct t_action port5_bst PROGMEM =   { (0<<7) + 5, BST_RUN*6, BST_ON, BST_OFF};
struct t_action port6_bst PROGMEM =   { (0<<7) + 6, BST_RUN*7, BST_ON, BST_OFF};
struct t_action port7_bst PROGMEM =   { (0<<7) + 7, BST_RUN*8, BST_ON, BST_OFF};
*/

//------------------------------------------------------------------------------
// This Routine copies the given action values to the corresponding array
// (loading the values is a semaphor operation, therefore we disable int)

/*
void perform_action(const struct t_action *pData)
  {
    unsigned char ctrl;
    unsigned char port;
    unsigned char mask; 

    ctrl = pgm_read_byte(&pData->status);

    port = ctrl & 0b0111;
    mask = 1 << port;
    cli();
    if (ctrl & (1<<7)) OUTPUT_PORT |= mask;
    else               OUTPUT_PORT &= ~mask;
    
    out_pwm[port].rest    = pgm_read_byte(&pData->rest);
    out_pwm[port].ontime  = pgm_read_byte(&pData->ontime);
    out_pwm[port].offtime = pgm_read_byte(&pData->offtime);
    if (ctrl & (1<<6)) out_pwm[port].rest    = pulsdelay;  // override default
    sei(); 
  }

*/



//==============================================================================
//
// Section 5
//
// Performing the Command

//------------------------------------------------------------------------------
// This Routine is called when myAdr is received


#if (PORT_ENABLED == TRUE)

void port_action(unsigned int Command, unsigned char Activate)
  {
    unsigned char myCommand, myTurnout;
    myCommand = Command & 0b00000111;
    myTurnout = myCommand >> 1;

    if (Command >= 8) return;                    // not our Address
    
    // checkout local alarm

    if (failed_turnout == myTurnout)   
      { // we got a command to the same turnout
        turn_led_off(); 
        if (feedback_method == ALARM) DCC_ACK_OFF;
        failed_turnout = 0x80;
      }

    if (Activate)                               //  activate coil?
      {
        
        disable_timer_interrupt(); 
        
        if (myCommand == 0)
          {
            out_pwm[0].rest = turnout[0].pulse_duration;         
            out_pwm[1].rest = 0;
            turnout[0].position = 0;         
            output(PB0,1);
            output(PB1,0);
            if (turnout[0].pulse_duration == 0)
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }
          }
        else if (myCommand == 1)
          {
            out_pwm[1].rest = turnout[0].pulse_duration;         
            out_pwm[0].rest = 0;         
            turnout[0].position = 1;         
            output(PB0,0);
            output(PB1,1);
            if (turnout[0].pulse_duration == 0) 
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }
          }
        else if (myCommand == 2)
          {
            out_pwm[2].rest = turnout[1].pulse_duration;         
            out_pwm[3].rest = 0;         
            turnout[1].position = 0;         
            output(PB2,1);
            output(PB3,0);
            if (turnout[1].pulse_duration == 0) 
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }
          }
        else if (myCommand == 3)
          {
            out_pwm[3].rest = turnout[1].pulse_duration;         
            out_pwm[2].rest = 0;         
            turnout[1].position = 1;         
            output(PB2,0);
            output(PB3,1);
            if (turnout[1].pulse_duration == 0)
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }

          }
        else if (myCommand == 4)
          {
            out_pwm[4].rest = turnout[2].pulse_duration;         
            out_pwm[5].rest = 0;         
            turnout[2].position = 0;         
            output(PB4,1);
            output(PB5,0);
            if (turnout[2].pulse_duration == 0)
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }
          }
        else if (myCommand == 5)
          {
            out_pwm[5].rest = turnout[2].pulse_duration;         
            out_pwm[4].rest = 0;         
            turnout[2].position = 1;         
            output(PB4,0);
            output(PB5,1);
            if (turnout[2].pulse_duration == 0)
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }
          }
        else if (myCommand == 6)
          {
            out_pwm[6].rest = turnout[3].pulse_duration;         
            out_pwm[7].rest = 0;         
            turnout[3].position = 0;         
            output(PB6,1);
            output(PB7,0);
            if (turnout[3].pulse_duration == 0)
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }
          }
        else // (myCommand == 7)
          {
            out_pwm[7].rest = turnout[3].pulse_duration;         
            out_pwm[6].rest = 0;         
            turnout[3].position = 1;         
            output(PB6,0);
            output(PB7,1);
            if (turnout[3].pulse_duration == 0)
              {
                PortState = PORTB;
                semaphor_set(C_DoSave);
              }
          }       
         
        enable_timer_interrupt(); 
      } // coil
    else
      {
        // coil off command - we also do a feedback
        disable_timer_interrupt(); 
        
        if (myCommand == 0)
          {
            out_pwm[0].rest = 0;         // stop every timer
            out_pwm[1].rest = 0;
            if (turnout[0].pulse_duration != 0)
              {                          // in pulse mode: allways turn off both coils
                output(PB0,0);
                output(PB1,0);
              }
          }
        else if (myCommand == 1)
          {
            out_pwm[1].rest = 0;         
            out_pwm[0].rest = 0;         
            if (turnout[0].pulse_duration != 0)
              {
                output(PB0,0);
                output(PB1,0);
              }
          }
        else if (myCommand == 2)
          {
            out_pwm[2].rest = 0;         
            out_pwm[3].rest = 0;         
            if (turnout[1].pulse_duration != 0)
              {
                output(PB2,0);
                output(PB3,0);
              }
          }
        else if (myCommand == 3)
          {
            out_pwm[3].rest = 0;         
            out_pwm[2].rest = 0;         
            if (turnout[1].pulse_duration != 0)
              {
                output(PB2,0);
                output(PB3,0);
              }
          }
        else if (myCommand == 4)
          {
            out_pwm[4].rest = 0;         
            out_pwm[5].rest = 0;         
            if (turnout[2].pulse_duration != 0)
              {
                output(PB4,0);
                output(PB5,0);
              }
          }
        else if (myCommand == 5)
          {
            out_pwm[5].rest = 0;         
            out_pwm[4].rest = 0;         
            if (turnout[2].pulse_duration != 0)
              {
                output(PB4,0);
                output(PB5,0);
              }
          }
        else if (myCommand == 6)
          {
            out_pwm[6].rest = 0;         
            out_pwm[7].rest = 0;         
            if (turnout[3].pulse_duration != 0)
              {
                output(PB6,0);
                output(PB7,0);
              }
          }
        else // (myCommand == 7)
          {
            out_pwm[7].rest = 0;         
            out_pwm[6].rest = 0;         
            if (turnout[3].pulse_duration != 0)
              {
                output(PB6,0);
                output(PB7,0);
              }
          }       
         
        enable_timer_interrupt(); 
        
        _mydelay_us(100);   // wait 100us to settle IO lines
        
        // feedback is communicated to the host depending on feedback_method:
        //      0:  not at all
        //      1:  positve feedback (announcing right position with a pulse)
        //      2:  negative feedback (ACK aktivated until the next turnout command)
        
        switch(feedback_method)
          {
            case NO_FB:
                if (check_position(myCommand) == 1) return;
                local_alarm(myTurnout);
                break;
            case POS_ACK:
                if (check_position(myCommand) == 1) 
                  {
                    activate_ACK(3);            // next preamble in 1.6ms over
                    return;
                  }
                else
                  {
                    local_alarm(myTurnout);
                  }
                break;
            case ALARM:
                if (check_position(myCommand) == 1) 
                  {
                    return;
                  }
                else
                  {
                    DCC_ACK_ON;                 // remote alarm (permanent)
                    local_alarm(myTurnout);
                  }
                break;
          }
      }
  }


// issue an alarm:
// parameter count: this turnout failed
//                  range: 0...3 -> turnout 1..4 

void local_alarm(unsigned char count)
  {
    failed_turnout = count;
    flash_led_fast(count+1);  // show the turnout
  }


//------------------------------------------------------------------------------------
// check_position tests the position of a turnout:
// depending on feedback mode of this turnout:
//      0:  no feedback, just a control of the previous command
//      1:  feedback through intrinsic coil end switches (like in roco line)
//      2:  feedback with extra indicator - PA0 is low, when in *red* position
//      3:  feedback with extra indicator - PA0 is high, when in *red* position
//
// PA0 may be PA2, PA4, PA6 depending on turnout
// PINA is aliased as FEEDBACK_IN
// PORTA is aliased as FEEDBACK_PULLUP  (note: this is write only)
//
// return: true (non zero) if position is okay
//         false if wrong


unsigned char check_position(unsigned char myCommand)
  {
    unsigned char myTurnout, myBits;

    myTurnout = myCommand >> 1;

    myBits = FEEDBACK_IN >> (myTurnout * 2);
    myBits &= 0b11;

    switch (turnout[myTurnout].feedback_mode)
      {
        case COMMAND_ONLY:
            if ((myCommand & 0b01) == 0b00)
              {
                // is it red?
                if (turnout[myTurnout].position == 0) return(1);    
              }
            else  
              {
                // is it green?
                if (turnout[myTurnout].position == 1) return(1);    
              }
            break;

        case END_SWITCH:
            if ((myCommand & 0b01) == 0b00)
              {
                // is it red?
                if (myBits == 0b10) return(1);
              }
            else
              {
                // is it green?
                if (myBits == 0b01) return(1);
              }
            break;

        case EXTERN_LOW:
            if ((myCommand & 0b01) == 0b00)
              {
                // is it red?
                if ((myBits & 0b01) == 0b00) return(1);
              }
            else
              {
                // is it green?
                if ((myBits & 0b01) == 0b01) return(1);
              }
            break;

        case EXTERN_HIGH:
            if ((myCommand & 0b01) == 0b00)
              {
                // is it red?
                if ((myBits & 0b01) == 0b01) return(1);
              }
            else
              {
                // is it green?
                if ((myBits & 0b01) == 0b00) return(1);
              }
            break;

      } 
    return(0);
  }

#endif





// Diese Funktion setzt die neuen Ziel-Bits fr DIMM
// beim Aufdimmen gibt es immer 400ms Verzgerung, damit ein Abdimmen vorher fertig wird.


void set_new_light_val(unsigned char valid, unsigned char pattern)
  {
    unsigned char port;
    unsigned char mask;
    unsigned char inv_pattern = ~pattern;

    mask = 1;
    for (port=0; port<8; port++)
      {
        if (pattern & mask & valid)
          {
            // turn on
            #if DIMM_DOWN_SAME_BULB == 1
                out_pwm[port].rest = DimmUpDelay;           // original:  DIMM_UP_DELAY / PWM_PERIOD;   
                CurrentTarget &= ~mask;
            #else
                if (CurrentTarget & mask)
                  {
                    // already on, nothing to do, keep burning 
                  }
                else
                  {
                    out_pwm[port].rest = DimmUpDelay;           // original:  DIMM_UP_DELAY / PWM_PERIOD;
                    CurrentTarget &= ~mask;
                  }
            #endif
          }
        if (inv_pattern & mask & valid)
          {
            // turn off
            out_pwm[port].rest    = 0;
            CurrentTarget &= ~mask;
          }
        mask = mask << 1;
      }
  }


void set_new_light_val_all(unsigned char valid, unsigned char pattern, unsigned char blink)
  {
    unsigned char port;
    unsigned char mask;
    
    mask = 1;
    for (port=0; port<8; port++)
      {
        if (valid & mask)
          {
            if (blink & mask)
              {
                if (pattern & mask)
                  {
                    CurrentTarget &= ~mask;                 // Target is off, this will turn on after .rest
                    out_pwm[port].rest = BlinkPeriod_2;
                    out_pwm[port].ontime = BlinkPeriod_2;
                    out_pwm[port].offtime = BlinkPeriod_2;
                  }
                else
                  {
                    CurrentTarget |= mask;                  // Target is on, this will turn off after .rest
                    out_pwm[port].rest = BlinkPeriod_2;
                    out_pwm[port].ontime = BlinkPeriod_2;
                    out_pwm[port].offtime = BlinkPeriod_2;
                  }
              }
            else   // static pattern
              {
                if (pattern & mask)
                  {
                    // turn on
                    #if DIMM_DOWN_SAME_BULB == 1
                        CurrentTarget &= ~mask;
                        out_pwm[port].rest = DimmUpDelay;           // original:  DIMM_UP_DELAY / PWM_PERIOD;   
                        out_pwm[port].ontime = 0;
                        out_pwm[port].offtime = 0;
                    #else
                        if (CurrentTarget & mask)
                          {
                            // already on, nothing to do, keep burning 
                          }
                        else
                          {
                            CurrentTarget &= ~mask;
                            out_pwm[port].rest = DimmUpDelay;           // original:  DIMM_UP_DELAY / PWM_PERIOD;   
                            out_pwm[port].ontime = 0;
                            out_pwm[port].offtime = 0;
                          }
                    #endif
                  }
                else
                  {
                    // turn off
                    CurrentTarget &= ~mask;
                    out_pwm[port].rest = 0;   
                    out_pwm[port].ontime = 0;
                    out_pwm[port].offtime = 0;
                  }
              }
          }
        mask = mask << 1;
      }
  }

//---------------------------------------------------------------------------------
// dimmer()
// this routine is called from main(), if C_Dimmstep is activated  
//

// Timing Engine
//
// Howto:    
// 1. Generelles Timing:
//    Diese Routine wird alle PWM_PERIOD aufgerufen. Es wird folgendes
//    geprft:
//    a) Wenn out_pwm[port].rest gleich 0: dann bleibt dieser Port unverndert.
//    b) out_pwm[port].rest wird decrementiert, wenn es dabei 0 wird, dann
//       wird ein Dimmvorgang in die andere Richtung eingeleitet.
//
// 2. Dimm-bergnge:
//    Je nach aktueller Richtung des Dimmvorgang (CurrentTarget) wird der aktuelle
//    Dimmwert erhht oder erniedrigt (z.Z. linear).
//    Die Dimmrampe ist unabhngig von den Zeiten, die bei ontime bzw. offtime
//    vorgegeben werden.
//    Wenn ein Ausgang ohne PWM durchschalten soll, dann mu sein Delta sehr
//    gro gewhlt werden! (>60)
  


void dimmer(void)
  {
    unsigned char port;
    unsigned char mask;
    unsigned char my_rest;

    mask = 1;
    for (port=0; port<8; port++)
      {
        my_rest = out_pwm[port].rest;       // use a local variable to force
        if (my_rest !=0)                    // compiler to tiny code
          { 
            if (--my_rest == 0)
              {
                 if (CurrentTarget & mask)
                  { // bit was on
                    my_rest = out_pwm[port].offtime;
                    CurrentTarget &= ~mask;
                  }
                else
                  {
                    my_rest = out_pwm[port].ontime;
                    CurrentTarget |= mask;
                  }
              }
            out_pwm[port].rest = my_rest;
          }


    	my_rest = out_pwm[port].dimm_val;
        if (CurrentTarget & mask)    
          {                                     // we are in phase A -> incr
            if (my_rest < out_pwm[port].light_A_val)  
              {
                // fehlt Sicherung gegen berlauf -> bei Daten aufpassen
                // delta darf nicht zu gro sein
				my_rest += out_pwm[port].delta_A;
              }
          }
        else
          {                                     // we are in phase B -> decr
            if (my_rest > out_pwm[port].light_B_val)
              {
                if (my_rest > out_pwm[port].delta_B)
					 my_rest -= out_pwm[port].delta_B;
                else my_rest = 0;
              }
          }
        out_pwm[port].dimm_val = my_rest;
        mask = mask << 1;                  // do *not* calc mask from port
      }

    // now process led

    if (led.rest != 0)
      {
        if (--led.rest == 0)
          {
             if (LED_STATE)
              {
                if (led.act_flash == led.flashes)
                  {
                    led.rest = led.pause;
                    led.act_flash = 0;
                  }
                else
                  {
                    led.rest = led.offtime;
                  }
                LED_OFF;
              }
            else
              {
                led.act_flash++;
                led.rest = led.ontime;
                LED_ON;
              }
          }
      }
// ???    enable_timer_interrupt();
  }

//==============================================================================
//

//==============================================================================
//
// Section 4
//
// Pattern for different signals (predefined in PROGMEM, read with pgm_read_byte)
//
// a) Vierfach rot grn
//
// Anschlu:
//
//  0baabbccdd     Anschlu:
//    |  ||  |---- Hauptsignal rot
//    |  || |----- Hauptsignal grn

unsigned char Signal_DB2_Pattern[8] PROGMEM =
  {
    0b00000001,          // hp0
    0b00000010,          // hp1
    0b00000100,
    0b00001000,
    0b00010000,
    0b00100000,
    0b01000000,
    0b10000000,
  };

unsigned char Signal_DB2_Valid[8] PROGMEM =
  {
    0b00000011,
    0b00000011,
    0b00001100,
    0b00001100,
    0b00110000,
    0b00110000,
    0b11000000,
    0b11000000,
  };


// b)  zwei unabhngige mehrbegriffige Vorsignale / Hauptsignale (a/b)
//
// Anschlu:
//
//  0baaaabbbb     Anschlu fr:
//    |  ||  |---- Vorsignal gelb 1;   Hauptsignal rot
//    |  || |----- Vorsignal gelb 2;   unused
//    |  |||------ Vorsignal grn 1;   Hauptsignal grn
//    |  ||------- Vorsignal grn 2;   Hauptsignal gelb

unsigned char Signal_DB3_Pattern[8] PROGMEM =
  {
    0b00000011,     // vr0         hp0
    0b00001100,     // vr1         hp2
    0b00000110,     // vr2         hp1
    0b00000000,     // unused      unused
    0b00110000,     // vr0         hp0
    0b11000000,     // vr1         hp2
    0b01100000,     // vr2         hp1
    0b00000000,     // unused      unused
  };

unsigned char Signal_DB3_Valid[8] PROGMEM =
  {
    0b00001111,
    0b00001111,
    0b00001111,
    0b00001111,
    0b11110000,
    0b11110000,
    0b11110000,
    0b11110000,
  };


//
// Hauptsignal und Vorsignal am gleichen Mast (mit Dunkeltastung)
//
// Anschlu:
//    0bvvvhhhhh
//      | ||   |---- HS rot 1
//      | ||  |----- HS rot 2
//      | || |------ HS gelb
//      | |||------- HS grn
//      | ||-------- HS wei
//      | |--------- VS vr0 (Dioden erforderlich, siehe Opendecoder)
//      ||---------- VS vr1
//      |----------- VS vr2

unsigned char Signal_DB4_Pattern[8] PROGMEM =
  {
    0b00000011,     // hp0
    0b00001000,     // hp1
    0b00001100,     // hp2
    0b00010001,     // sh1
    0b00100000,     // vr0
    0b01000000,     // vr1
    0b10000000,     // vr2
    0b00000000,     // Dunkeltastung
  };

unsigned char Signal_DB4_Valid[8] PROGMEM =
  {
    0b00011111,
    0b00011111,
    0b00011111,
    0b00011111,
    0b11100000,
    0b11100000,
    0b11100000,
    0b11100000,
  };


//
// Alex Shepherd - 4 aspects binary coded (red, green, both on, none)
// Mode 6:

unsigned char Signal_comb_Pattern[16] PROGMEM =
  {
    0b00000001,    // only red    
    0b00000010,    // only green 
    0b00000011,    // red and green
    0b00000000,    // dark 
    0b00000100,     
    0b00001000,     
    0b00001100,     
    0b00000000,  
    0b00010000,    // only red    
    0b00100000,    // only green 
    0b00110000,    // red and green
    0b00000000,    // dark 
    0b01000000,     
    0b10000000,     
    0b11000000,     
    0b00000000,     
  };

unsigned char Signal_comb_Valid[16] PROGMEM =
  {
    0b00000011,
    0b00000011,
    0b00000011,
    0b00000011,
    0b00001100,
    0b00001100,
    0b00001100,
    0b00001100,
    0b00110000,
    0b00110000,
    0b00110000,
    0b00110000,
    0b11000000,
    0b11000000,
    0b11000000,
    0b11000000,
  };


unsigned char Signal_bin_Pattern[16] PROGMEM =
  {
    0b00000001,    // only red    
    0b00000000,    // red off
    0b00000010,    // green
    0b00000000,    // off
    0b00000100,     
    0b00000000,     
    0b00001000,     
    0b00000000,  
    0b00010000,    // red on   
    0b00000000,    // off
    0b00100000,    // green on
    0b00000000,    // off
    0b01000000,     
    0b00000000,     
    0b10000000,     
    0b00000000,     
  };

unsigned char Signal_bin_Valid[16] PROGMEM =
  {
    0b00000001,
    0b00000001,
    0b00000010,
    0b00000010,
    0b00000100,
    0b00000100,
    0b00001000,
    0b00001000,
    0b00010000,
    0b00010000,
    0b00100000,
    0b00100000,
    0b01000000,
    0b01000000,
    0b10000000,
    0b10000000,
  };


// signalling according to belgain signals
    // there are 4 lights and 7 aspects. So we can two signals, using 16 addresses
    // map
    // > > addr +0, r: single red                 1. Signal
    // > > addr +0, g: single green
    // > > addr +1, r: yellow yellow
    // > > addr +1, g: green yellow
    // > > addr +2, r: single red flashing
    // > > addr +2, g: green flashing
    // > > addr +3, r: yellow yellow flashing
    // > > addr +3, g: -
    // > > addr +4, r: single red                 2. Signal
    // > > addr +4, g: single green
    // > > addr +5, r: yellow yellow
    // > > addr +5, g: green yellow
    // > > addr +6, r: single red flashing
    // > > addr +6, g: green flashing
    // > > addr +7, r: yellow yellow flashing
    // > > addr +7, g: -

    // connect as follows:
    // Port 0  4: green
    // Port 1  5: red
    // Port 2  6: yellow (in row)
    // Port 3  7: yellow (aside)




struct s_pattern
    { unsigned char valid;      // one entry for each aspect, 1 marks the affected outputs
      unsigned char pattern;    // one entry for each aspect, target pattern
      unsigned char blink;     // one entry for each aspect, target pattern
    };

struct s_pattern Signal_Belgian[16] PROGMEM =
  { { 0b00001111, 0b00100010, 0b00000000},   // single red
    { 0b00001111, 0b00010001, 0b00000000},   // single green
    { 0b00001111, 0b11001100, 0b00000000},   // yellow yellow
    { 0b00001111, 0b10011001, 0b00000000},   // green yellow
    { 0b00001111, 0b00100010, 0b00100010},   // single red flashing
    { 0b00001111, 0b00010001, 0b00010001},   // single green flashing
    { 0b00001111, 0b11001100, 0b11001100},   // yellow yellow flashing
    { 0b00001111, 0b00000000, 0b00000000},   // all off
    { 0b11110000, 0b00100010, 0b00000000},   // single red
    { 0b11110000, 0b00010001, 0b00000000},   // single green
    { 0b11110000, 0b11001100, 0b00000000},   // yellow yellow
    { 0b11110000, 0b10011001, 0b00000000},   // green yellow
    { 0b11110000, 0b00100010, 0b00100010},   // single red flashing
    { 0b11110000, 0b00010001, 0b00010001},   // single green flashing
    { 0b11110000, 0b11001100, 0b11001100},   // yellow yellow flashing
    { 0b11110000, 0b00000000, 0b00000000},   // all off
  }; 


//==============================================================================
//
// Section 5
//
void rail_crossing(unsigned char myCommand)
  {
  {                                            // Wechselblinker mit Stopmagnet
         // Port 0, 1: Andreaskreuze -> Wechselblinken 
         // Port 2: schnelles Blinkausgang (Lokfhrer)
         // Port 3: Schalten des Stopmagneten
         if (myCommand == 0) // = schalte alles ab
           {
             #define BUE_AK_BL  ( 1000000L / PWM_PERIOD)   // Blinken Andreaskreuz
             #define BUE_LF_BL  ( 1000000L / PWM_PERIOD)   // Blinken Lokfhrer
         
                  set_new_light_val(0x07, 0x00);          // turn all off
                  set_target(3,1); // Magnet;
                  out_pwm[3].rest    = 2* BUE_AK_BL;  // StopMagnet nach 2 Blinkphasen aus
                  out_pwm[3].offtime = 0;                   
				  // Hinweis: braucht groes Delta - wird beim Einstellen des Modes gemacht!
                  hauptsignal_state = myCommand;
           }
         else if (myCommand == 1) // = schalte beide ein, Phasenversatz
           {
                  set_target(3,0); // Magnet
                  set_target(2,0); // gn;
                  set_target(1,0); // ye;
                  set_target(0,0); // rd;
                  
                  out_pwm[0].rest    = 2*BUE_AK_BL/2;
                  out_pwm[0].offtime = BUE_AK_BL/2;
                  out_pwm[0].ontime  = BUE_AK_BL/2;
                
                  
                  out_pwm[1].rest    = BUE_AK_BL/2;
                  out_pwm[1].offtime = BUE_AK_BL/2;
                  out_pwm[1].ontime  = BUE_AK_BL/2;
                
                  out_pwm[2].rest    = BUE_LF_BL/2;
                  out_pwm[2].offtime = BUE_LF_BL/2;
                  out_pwm[2].ontime  = BUE_LF_BL/2;
                
                  out_pwm[3].rest    = 2* BUE_AK_BL;  // StopMagnet nach 2 Blinkphasen an
                  out_pwm[3].ontime = 0;                   
				  // Hinweis: braucht groes Delta - wird beim Einstellen des Modes gemacht!
                  hauptsignal_state = myCommand;           
           }
         else if (myCommand == 2) // unused
           {
          }                     
         else if (myCommand == 3) // unused
           {
           }
         else if (myCommand == 4) 
           {          
           // = schalte alles ab (2. Gruppe)
                  
                  set_new_light_val(0x70, 0x00);
                  set_target(7,1); // Magnet
                  out_pwm[7].rest    = 2* BUE_AK_BL;  // StopMagnet nach 2 Blinkphasen aus
                  out_pwm[7].offtime = 0;                   
				  // Hinweis: braucht groes Delta - wird beim Einstellen des Modes gemacht!
                  vorsignal_state = myCommand;
           }
         else if (myCommand == 5) // = schalte beide ein, Phasenversatz
           {
                  set_target(7,0); // Magnet
                  set_target(6,0); // gn;
                  set_target(5,0); // ye;
                  set_target(4,0); // rd;
                  
                  out_pwm[4].rest    = 2*BUE_AK_BL/2;
                  out_pwm[4].offtime = BUE_AK_BL/2;
                  out_pwm[4].ontime  = BUE_AK_BL/2;
                
                  out_pwm[5].rest    = BUE_AK_BL/2;
                  out_pwm[5].offtime = BUE_AK_BL/2;
                  out_pwm[5].ontime  = BUE_AK_BL/2;
                
                  out_pwm[6].rest    = BUE_LF_BL/2;
                  out_pwm[6].offtime = BUE_LF_BL/2;
                  out_pwm[6].ontime  = BUE_LF_BL/2;
                
                  out_pwm[7].rest    = 2* BUE_AK_BL;  // StopMagnet nach 2 Blinkphasen
                  out_pwm[7].ontime = 0;                   
				  // Hinweis: braucht groes Delta - wird beim Einstellen des Modes gemacht!
                  vorsignal_state = myCommand;
           }
         else if (myCommand == 6) // unused
           {
                
           }                     
         else if (myCommand == 7) // unused
           {
                
           }
      }
  }


void light_chain_fwd(unsigned char myCommand)
  {
    // command 0: all off
    //         1: single bulb forward (looping)
    //         2: dual bulb forward (looping)
    //         3: triple bulb (looping)
    unsigned char port;
    unsigned char short_dwell;
    unsigned char long_dwell;
    unsigned char ontime;

    short_dwell = eeprom_read_byte(&CV.StrokePeriod);
    if (short_dwell == 0) short_dwell = 1;

    ontime = eeprom_read_byte(&CV.Ontime);
    if (ontime == 0) ontime = 1;

    long_dwell = eeprom_read_byte(&CV.IntervalPeriod);
    // if (long_dwell <= short_dwell) long_dwell = short_dwell+1;
    if (long_dwell <= ontime) long_dwell = ontime+1;
    
    switch(myCommand)
      {
        case 0:
            set_new_light_val(0b11111111, 0);
            hauptsignal_state = myCommand;           
            break;
        case 1:
            CurrentTarget = 0;
            for (port=0; port<8; port++)
               {
                 out_pwm[port].rest = (1 + port) * short_dwell;
                 out_pwm[port].ontime = ontime;
                 out_pwm[port].offtime = long_dwell - ontime;
               }
            hauptsignal_state = myCommand;           
            break;
        case 2:
            ontime = 2 * ontime;
            if (long_dwell <= ontime) long_dwell = ontime+1;
            CurrentTarget = 0;
            for (port=0; port<8; port++)
               {
                 out_pwm[port].rest = (1 + port) * short_dwell;
                 out_pwm[port].ontime = ontime;
                 out_pwm[port].offtime = long_dwell - ontime;
               }
            hauptsignal_state = myCommand;           
            break;
        case 3:
            ontime = 3 * ontime;
            if (long_dwell <= ontime) long_dwell = ontime+1;
            CurrentTarget = 0;
            for (port=0; port<8; port++)
               {
                 out_pwm[port].rest = (1 + port) * short_dwell;
                 out_pwm[port].ontime = ontime;
                 out_pwm[port].offtime = long_dwell - ontime;
               }
            hauptsignal_state = myCommand;           
            break;
        default:
            break;
      }
  }

void light_column(unsigned char myCommand)
  {
    // command 0: all off
    //         1: up - down
    unsigned char port;
    unsigned char short_dwell;
    unsigned char long_dwell;
    

    short_dwell = eeprom_read_byte(&CV.StrokePeriod);
    if (short_dwell == 0) short_dwell = 1;

    long_dwell = eeprom_read_byte(&CV.IntervalPeriod);
    // if (long_dwell <= short_dwell) long_dwell = short_dwell+1;
   
    switch(myCommand)
      {
        case 0:
            set_new_light_val(0b11111111, 0);
            hauptsignal_state = myCommand;           
            break;
        case 1:
            CurrentTarget = 0;
            for (port=0; port<8; port++)
               {
                 out_pwm[port].rest = (1 + port) * short_dwell;
                 out_pwm[port].ontime = (8 - port) * 2;
                 out_pwm[port].offtime = long_dwell;
               }
            hauptsignal_state = myCommand;           
            break;
       default:
            break;
      }
  }



void traffic_light(unsigned char myCommand)
  {
    // Traffic Light (Ampel)
    // Ampel
    // LED0:RT1 xxxxxxxx_____________________________xxxxxxxxxxxxxxxxx
    // LED1:GE1 ______xx___________________________xx_________________
    // LED2:GN1 ________xxxxxxxxxxxxxxxxxxxxxxxxxxx___________________
    // LED3:STOP1 xxxxxxx___________________________xxxxxxxxxxxxxxxxxx

    #define AMP_YE  (1000000L / PWM_PERIOD)   // Yellow 1 s
    #define AMP_BL  ( 500000L / PWM_PERIOD)   // Blinken
    if (myCommand == 0) // = schalte auf rot
      {
        set_target(3,0); // Magnet
        set_target(2,0); // gn;
        set_target(1,1); // ye;
        set_target(0,0); // rd;

        out_pwm[0].rest    = AMP_YE;
        // out_pwm[0].ontime  = 0;

        out_pwm[1].rest    = AMP_YE;
        out_pwm[1].offtime = 0;

        out_pwm[2].rest    = 0;

        out_pwm[3].rest    = AMP_YE / 2;  // StopMagnet nach halber Gelb ein
        // out_pwm[3].ontime = 0;
        // Hinweis: braucht groes Delta - wird beim Einstellen des Modes gemacht!
        hauptsignal_state = myCommand;           
      }
    else if (myCommand == 1) // = schalte auf grn
      {
              set_target(3,1); // Magnet
              set_target(2,0); // gn;
              set_target(1,1); // ye;
              set_target(0,1); // rd;

              out_pwm[0].rest    = AMP_YE;
              //out_pwm[0].offtime  = 0;

              out_pwm[1].rest    = AMP_YE;
              out_pwm[1].offtime = 0;

              out_pwm[2].rest    = AMP_YE;
              //out_pwm[2].ontime = 0;

              out_pwm[3].rest    = 3*AMP_YE / 2;   // Magnet nach 1,5 gelb aus
              //out_pwm[3].offtime = 0;
        hauptsignal_state = myCommand;           
       }
     else if (myCommand == 2) // = alles aus
       {
              set_new_light_val(0x0f, 0x00);

              /*
              CurrentTarget = (0<<3) | (0<<2) | (0<<1) | (0<<0);
                                  //              Stop     green    gelb   rot
              out_pwm[1].rest    = 0;
              out_pwm[0].rest    = 0;
              out_pwm[2].rest    = 0;
              out_pwm[3].rest    = 0;
              */
        hauptsignal_state = myCommand;           
      }
    else if (myCommand == 3) // blinken
      {
        set_target(3,0); // Magnet
        set_target(2,0); // gn;
        set_target(1,1); // ye;
        set_target(0,0); // rd;

        out_pwm[1].rest    = AMP_YE;
        out_pwm[1].ontime  = AMP_YE;
        out_pwm[1].offtime = AMP_YE;

        out_pwm[0].rest    = 0;
        out_pwm[2].rest    = 0;
        out_pwm[3].rest    = 0;
        hauptsignal_state = myCommand;           
      }
    else if (myCommand == 4) // = schalte auf rot
      {
              set_target(7,0); // Magnet
              set_target(6,0); // gn;
              set_target(5,1); // ye;
              set_target(4,0); // rd;

              out_pwm[4].rest    = AMP_YE;
              out_pwm[4].ontime  = 0;

              out_pwm[5].rest    = AMP_YE;
              out_pwm[5].offtime = 0;

              out_pwm[6].rest    = 0;

              out_pwm[7].rest    = AMP_YE / 2;
              out_pwm[7].ontime = 0;
        vorsignal_state = myCommand;           
      }
    else if (myCommand == 5) // = schalte auf grn
      {
        set_target(7,1); // Magnet
        set_target(6,0); // gn;
        set_target(5,1); // ye;
        set_target(4,1); // rd;

        out_pwm[4].rest    = AMP_YE;
        out_pwm[4].offtime  = 0;

        out_pwm[5].rest    = AMP_YE;
        out_pwm[5].offtime = 0;

        out_pwm[6].rest    = AMP_YE;
        out_pwm[6].ontime = 0;

        out_pwm[7].rest    = 3*AMP_YE / 2;
        out_pwm[7].offtime = 0;
        vorsignal_state = myCommand;           
      }
    else if (myCommand == 6) // = alles aus
      {
        set_new_light_val(0xf0, 0x00);

        /*
        CurrentTarget = (0<<7) | (0<<6) | (0<<5) | (0<<4);
                          //              Stop     green    gelb   rot
        out_pwm[5].rest    = 0;
        out_pwm[4].rest    = 0;
        out_pwm[6].rest    = 0;
        out_pwm[7].rest    = 0;
        */
        vorsignal_state = myCommand;           
       }
    else if (myCommand == 7) // blinken
      {
        set_target(7,0); // Magnet
        set_target(6,0); // gn;
        set_target(5,1); // ye;
        set_target(4,0); // rd;

        out_pwm[5].rest    = AMP_YE;
        out_pwm[5].ontime  = AMP_YE;
        out_pwm[5].offtime = AMP_YE;

        out_pwm[4].rest    = 0;
        out_pwm[6].rest    = 0;
        out_pwm[7].rest    = 0;
        vorsignal_state = myCommand;           
      }
  }

void set_signal_aspects(myCommand)
  {
    switch (eeprom_read_byte(&CV.SignalType))
      {
        case 0:                         // 00 = User Defined
                if (myCommand < eeprom_read_byte(&CV.NoOfAspects))
                    set_new_light_val_all(
                        eeprom_read_byte(&CV.UDEF[myCommand].Valid),
                        eeprom_read_byte(&CV.UDEF[myCommand].Pattern),
                        eeprom_read_byte(&CV.UDEF[myCommand].Blink)    );
                break;
        default:  
        case 1:                         // Vier zweibegriffige Signale
                myCommand = myCommand & 0x07;
                set_new_light_val(
                    pgm_read_byte(&Signal_DB2_Valid[myCommand]),
                    pgm_read_byte(&Signal_DB2_Pattern[myCommand])  );
                break;
        case 2:                         // Zwei dreibegriffige Signale (DB)
                myCommand = myCommand & 0x07;
                set_new_light_val(
                    pgm_read_byte(&Signal_DB3_Valid[myCommand]),
                    pgm_read_byte(&Signal_DB3_Pattern[myCommand])  );
                break;
        case 3:                     // Einfahrsignal + Vorsignal getrennt
                myCommand = myCommand & 0x07;
                set_new_light_val(
                    pgm_read_byte(&Signal_DB4_Valid[myCommand]),
                    pgm_read_byte(&Signal_DB4_Pattern[myCommand])  );
                break;
        case 4:                     // Einfahrsignal + Vorsignal mit Dunkeltastung
                if (myCommand <= 3)
                  {
                    if ((hauptsignal_state == 0)||(hauptsignal_state == 3)) 
                      {
                        // VS war aus, wieder anmachen
                        set_new_light_val(
                            pgm_read_byte(&Signal_DB4_Valid[vorsignal_state]),
                            pgm_read_byte(&Signal_DB4_Pattern[vorsignal_state])  );  
                      }
                    hauptsignal_state = myCommand;
                    set_new_light_val(
                        pgm_read_byte(&Signal_DB4_Valid[hauptsignal_state]),
                        pgm_read_byte(&Signal_DB4_Pattern[hauptsignal_state])  ); 
                  }
                else if (myCommand <= 7)
                  {
                    vorsignal_state = myCommand;
                    set_new_light_val(
                        pgm_read_byte(&Signal_DB4_Valid[vorsignal_state]), 
                        pgm_read_byte(&Signal_DB4_Pattern[vorsignal_state])  );
                  }

                if ((hauptsignal_state == 0)||(hauptsignal_state == 3)) 
                  {                                        // mit Dunkeltastung
                    // einfach drber schreiben
                    set_new_light_val(
                        pgm_read_byte(&Signal_DB4_Valid[7]),
                        pgm_read_byte(&Signal_DB4_Pattern[7])     );
                  }
                break;
        case 5:                         // 16 addresses, dimmed, individual control
                set_new_light_val(
                    pgm_read_byte(&Signal_bin_Valid[myCommand]),
                    pgm_read_byte(&Signal_bin_Pattern[myCommand])  );
                break;

        case 6:                         // up to 16 addresses, dimmed, logic output
                set_new_light_val(
                    pgm_read_byte(&Signal_comb_Valid[myCommand]),
                    pgm_read_byte(&Signal_comb_Pattern[myCommand])  );
                break;
        case 7:
                set_new_light_val_all(
                    pgm_read_byte(&Signal_Belgian[myCommand].valid),
                    pgm_read_byte(&Signal_Belgian[myCommand].pattern),
                    pgm_read_byte(&Signal_Belgian[myCommand].blink)  );
                break;
      }
  }


//------------------------------------------------------------------------------
// This Routine is called when myAdr is received



void action(unsigned int Command, unsigned char Activate)
  {
    unsigned char myCommand;
    unsigned char myLightMode;
    
    if (Command >= 16) return;                                 // not for me

    myCommand = Command;
      
    myLightMode = eeprom_read_byte(&CV.LightMode);

    if (myLightMode == 0)
      {
        port_action(Command, Activate);       // 00 = 4 pairs, switched, 8 addresses (traditional)
      }                                                          //      (duration according to T_on_xx)
    else
      {
        if (Activate)
          {
            switch (myLightMode)
              {
                case  0:break;
                default:
                case  1: // 01 = 8 addresses, dimmed, static states
                        if (myCommand < 8)
                           set_signal_aspects(myCommand);
                        break;
                case  2: // 02 = 16 addresses, dimmed, static states
                        set_signal_aspects(myCommand);
                        break;
                
                case  8: // 08 = traffic lights (Ampel) 8 addresses
                        traffic_light(myCommand);
                        break;
                case  9: // 09 = rail crossing (Andreaskreuz) 8 addresses
                        rail_crossing(myCommand);
                        break;
                case 10: // 10 = Flasher
                        break;
                case 16: // 16 = Lichtsule (aus, langsam, schnell)
                        light_column(myCommand);
                        break;
                case 17: // 17 = Lichterkette vorwrts (aus, eine langsam, zwei langsam, drei,  schnell)
                        light_chain_fwd(myCommand);
                        break;
                case 18: // 18 = Lichterkette rckwrts (aus, eine langsam, zwei langsam, drei,  schnell)
                        break;
                case 19: // 19 = Baustellenblitzer
                        break;
              }   
          }
        else
          {
            // we are always true
            // feedback is communicated to the host depending on feedback_method:
            //      0:  not at all
            //      1:  positve feedback (announcing right position with a pulse)
            //      2:  negative feedback (ACK aktivated until the next turnout command)
            switch(feedback_method)
              {
                case NO_FB:
                    break;
                case POS_ACK:
                    activate_ACK(3);            // next preamble in 1.6ms over
                    break;
                case ALARM:
                    break;
              }
          } 
        semaphor_set(C_DoSave);
      }
  }


void simple_preset(unsigned char command)
  {
    // during programming with tracer, we do a mode select depending on the received
    // command.
    
    // two different setups:
    // with Jumper: Signal States
    // without Jumper: Turnout Decoder

    unsigned char pulsdelay;

    if (JUMPER_FITTED)
      {
        switch(command)
          {
            default:
            case 0:
                                // Vier zweibegriffige Signale
                eeprom_write_byte(&CV.LightMode, 1);
                eeprom_write_byte(&CV.SignalType, 1);
                break;
            case 1:             // Zwei dreibegriffige Signale oder Vorsignale (DB),
                                // wobei jede LED einen eigenen Ausgang hat (keine Dioden erforderlich)
                eeprom_write_byte(&CV.LightMode, 1);
                eeprom_write_byte(&CV.SignalType, 2);
                break;
            case 2:             // Einfahrsignal und Vorsignal
                eeprom_write_byte(&CV.LightMode, 1);
                eeprom_write_byte(&CV.SignalType, 3);
                break;
            case 3:             // Einfahrsignal und Vorsignal, mit Dunkeltastung des Vorsignals
                eeprom_write_byte(&CV.LightMode, 1);
                eeprom_write_byte(&CV.SignalType, 4);
                break;
            case 4:             // Wechselblinker (fr Bahnbergang, Fugngerberweg),
                                // zustzliche Ansteuerung eines Stopmagneten fr Faller Car.
                eeprom_write_byte(&CV.LightMode, 9);
                break;
            case 5:             // reserved (Einzelschalter?)
                eeprom_write_byte(&CV.LightMode, 2);
                eeprom_write_byte(&CV.SignalType, 5);
                break;
            case 6:             // 4 vierbegriffige Signale (rot, grn, gelb, aus), binr codiert.
                eeprom_write_byte(&CV.LightMode, 2);
                eeprom_write_byte(&CV.SignalType, 6);
                break;
            case 7:             // Ampel, mit Ansteuerung Stopmagnet fr Faller Car
                eeprom_write_byte(&CV.LightMode, 9);
                break;
          }

        eeprom_write_byte(&CV.FBM_F1, 0);    // 0 = COMMAND_ONLY
        eeprom_write_byte(&CV.FBM_F2, 0);    // 0 = COMMAND_ONLY
        eeprom_write_byte(&CV.FBM_F3, 0);    // 0 = COMMAND_ONLY
        eeprom_write_byte(&CV.FBM_F4, 0);    // 0 = COMMAND_ONLY

        do {} while (!eeprom_is_ready());    // wait for write to complete      
      }
    else
      {
        eeprom_write_byte(&CV.LightMode, 0);                            // standard decoder
      
        if (command == 0)      pulsdelay =  200000L / TICK_PERIOD; 
        else if (command == 1) pulsdelay =  500000L / TICK_PERIOD; 
        else if (command == 2) pulsdelay = 1000000L / TICK_PERIOD; 
        else if (command == 3) pulsdelay = 2000000L / TICK_PERIOD; 
        else                     pulsdelay = 0;   // permanent; 
   
        eeprom_write_byte(&CV.T_on_F1, pulsdelay);   
        eeprom_write_byte(&CV.T_on_F2, pulsdelay);   
        eeprom_write_byte(&CV.T_on_F3, pulsdelay);   
        eeprom_write_byte(&CV.T_on_F4, pulsdelay);   
 
        if (pulsdelay == 0)
          {
            eeprom_write_byte(&CV.LastState, 0x55);
            eeprom_write_byte(&CV.FBM_F1, 0);    // 0 = COMMAND_ONLY
            eeprom_write_byte(&CV.FBM_F2, 0);    // 0 = COMMAND_ONLY
            eeprom_write_byte(&CV.FBM_F3, 0);    // 0 = COMMAND_ONLY
            eeprom_write_byte(&CV.FBM_F4, 0);    // 0 = COMMAND_ONLY
          }
        else
          {
            eeprom_write_byte(&CV.LastState, 0x00);
            eeprom_write_byte(&CV.FBM_F1, 1);    // 1 = END_SWITCH
            eeprom_write_byte(&CV.FBM_F2, 1);    // 1 = END_SWITCH
            eeprom_write_byte(&CV.FBM_F3, 1);    // 1 = END_SWITCH
            eeprom_write_byte(&CV.FBM_F4, 1);    // 1 = END_SWITCH
          }
                    
        do {} while (!eeprom_is_ready());    // wait for write to complete
      }
    

  }




#if (PRBS_CODE == 1)
// linear feedback shift register (prbs)
//
//      |---|    |---|    |---|    |---|    |---|    |---|    |---|    |---|
//    ->| 0 |--->| 1 |-o->| 2 |-o->| 3 |-o->| 4 |--->| 5 |--->| 6 |--->| 7 |--o--->
//   |  |---|    |---| |  |---| |  |---| |  |---|    |---|    |---|    |---|  |
//   |                 |        |        |                                    |
//    <--------------- + <----- + <----- + <----------------------------------
//
unsigned char prbs8(unsigned char seed)
  {
    unsigned char new_rnd;

    new_rnd = seed;                // copy bit 1
    new_rnd = new_rnd << 1;
    new_rnd = new_rnd ^ seed;      // xor bit 2
    new_rnd = new_rnd << 1;
    new_rnd = new_rnd ^ seed;      // xor bit 3
    new_rnd = new_rnd << 4;
    new_rnd = new_rnd ^ seed;      // xor bit 7

    // now put this bit to seed's lsb
    new_rnd = new_rnd >> 7;
    seed = seed << 1;
    new_rnd = new_rnd + seed; 
    
    return(new_rnd);
  }
#endif



