//------------------------------------------------------------------------
//
// OpenDCC - OpenDecoder2
//
// Copyright (c) 2006, 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-02-14 V0.1 kw copied from opendecoder.c
//                               split up the code in modules
//                               added routines for CV-Programming
//            2007-02-25 V0.2 kw corrected some items with Decoding
//            2007-04-10 V0.3 kw moved Timerinit to port_engine.c
//            2007-04-27 V0.4 kw all-in-one software
//            2007-05-31 V0.5 kw added Hardware 3, added Jumperquery
//                               for programming
//            2007-08-05 V0.6 kw change to CV struct 
//            2007-08-18 V0.7 kw added 5 flashes  if unprogrammed
//            2008-05-26 V0.8 kw added save of CurrentTarget if SignalDecoder
//            2008-05-26 V0.9 kw added CheckSave - saves CurrentTarget after 3s
//                               and hauptsignal_state + vorsignal_state
//
// 2do: manual adjustment, tracers
//
//------------------------------------------------------------------------
//
// purpose:   flexible general purpose decoder for dcc
//            here: init, mainloop, globals
//
//
// ====== >>>> config.h  is the central definition file for the project
//
//
// content:   A DCC-Decoder for ATmega8515 and other AVR
//
//             1. Defines and variable definitions
//             2. Init
//             3. DCC Parser
//             5. MAIN: analyze command, call the action, do programming
//             6. Test and Simulation
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"              // general definitions the decoder, cv's
#include "hardware.h"            // port definitions for target
#include "dcc_receiver.h"        // receiver for dcc
#include "dcc_decode.h"          // decode dcc and handle CV access
#include "dimm_engine.h"         // handling of output and feedback,
                                 // LED control
// #include "dmxout.h"              // dmx
// #include "servo.h"               // servo
#include "keyboard.h"

#include "main.h"

#define SIMULATION  0               // 0: real application
                                    // 1: test receive routine
                                    // 2: test timing engine
                                    // 3: test action


//-----------------------------------------------------------------------------------
//
// Security Checks
//

#ifndef PROCESSOR
 #warning: there is no target processor defined - see hardware.h
 #warning: recommended: ATmega8515 (8k Flash)
#endif
//-----------------------------------------------------------------------------------
//
// Global Data
//


unsigned char PortState;            // this is the state to be saved


//--------------------------------------------------------------------------------------

#if (TARGET_HARDWARE == OPENDECODER2)

void init_main(void)
  {
    DDRA  = 0;                  // PORTA: inputs
    DDRB  = 0xFF;               // PortB: All Bits as Output
    DDRC  = 0;                  // PORTC: inputs
    DDRD  = (0<<RS485_RX)       // input
          | (1<<RS485_TX)       // output
          | (0<<DCCIN) 
          | (0<<JUMPER)
          | (0<<PROGTASTER)   
          | (1<<SERVO1)         // output (OC1A)
          | (1<<LED)            // output, 1 turns on LED
          | (1<<DCC_ACK);       // output, sending 1 makes an ACK
    DDRE  = (1<<DMXDIR)         // output
          | (1<<SERVO2);        // output (OC1B)
       
    PORTA = 0;                  // feedback: all pull down
    PORTB = 0;                  // output: all off     
    PORTD = (1<<RS485_RX)       // 1 = pullup
          | (0<<RS485_TX)       // 
          | (1<<DCCIN)          // 1 = pullup
          | (1<<JUMPER)         // 1 = pullup
          | (1<<PROGTASTER)     // 1 = pullup 
          | (0<<SERVO1)         // output (OC1A)
          | (0<<LED)            // LED off
          | (0<<DCC_ACK);       // ACK off
    PORTE = (1<<DMXDIR)         // 0=receive, 1=transmit
          | (0<<SERVO2);        // output (OC1B)
  }

#elif (TARGET_HARDWARE == OPENDECODER3) 

void init_main(void)
  {
    DDRA  = 0;                  // PORTA: nothing -> inputs
    DDRB  = (1<<LED)            // output, 1 turns on LED)
          | (1<<DMXDIR)         // 0=receive, 1=transmit
          | (0<<RS485_RX)       // input
          | (1<<RS485_TX);      // output
    DDRC  = (1<<RELAIS1)        // PORTC: 4* Relais
          | (1<<RELAIS2) 
          | (1<<RELAIS3) 
          | (1<<RELAIS4)
          | (0<<KEY1)           // input
          | (0<<KEY2)           // input
          | (0<<KEY3)           // input
          | (0<<KEY4);          // input
 
    DDRD  = (0<<LN_IN)          // input
          | (1<<RAILCOM)        // output
          | (0<<DCCIN) 
          | (0<<JUMPER2)
          | (0<<PROGTASTER)   
          | (1<<SERVO1)         // output (OC1A)
          | (1<<RAILCOM_ON)
          | (1<<DCC_ACK);       // output, sending 1 makes an ACK

    DDRE  = (0<<JUMPER1)        // input
          | (1<<SERVO2);        // output (OC1B)
       
    PORTA = 0;                  // nothing
    PORTB = (0<<LED)            // output, 1 turns on LED)
          | (1<<DMXDIR)         // 0=receive, 1=transmit
          | (1<<RS485_RX)       // input -> pullup
          | (0<<RS485_TX);      // output
    PORTC = (0<<RELAIS1)        // PORTC: 4* Relais
          | (0<<RELAIS2) 
          | (0<<RELAIS3) 
          | (0<<RELAIS4)
          | (1<<KEY1)           // input -> pullup
          | (1<<KEY2)           // input -> pullup
          | (1<<KEY3)           // input -> pullup
          | (1<<KEY4);          // input -> pullup
 
    PORTD = (1<<LN_IN)          // input -> pullup
          | (0<<RAILCOM)        // output
          | (1<<DCCIN) 
          | (1<<JUMPER2)
          | (1<<PROGTASTER)   
          | (0<<SERVO1)         // output (OC1A)
          | (0<<RAILCOM_ON)
          | (0<<DCC_ACK);       // output, sending 1 makes an ACK

    PORTE = (1<<JUMPER1)        // input
          | (0<<SERVO2);        // output (OC1B)
  
  }
#endif

//--------------------------------------------------------------------------------------------------
//
// here we check out whether we should save a current state of the decoder
//
// CurrentTarget should no be saved immediately after a command, so we do some timeout 

#define SAVE_TIMEOUT  (3000000L / TICK_PERIOD)
#if (SAVE_TIMEOUT == 0)
 #define SAVE_TIMEOUT  1
#endif

#if (SAVE_TIMEOUT > 240)
 #warning SAVE_TIMEOUT  too large
#endif

unsigned char CheckSaveState;
signed char CheckSave_timerval;
    

// saves:
unsigned char RAM_LastState;
//  hauptsignal_state;
//  vorsignal_state;


void CheckSave(void)
  {
    switch (CheckSaveState)
      {
        default:
        case 0:
            RAM_LastState = eeprom_read_byte(&CV.LastState);
            hauptsignal_state = eeprom_read_byte(&CV.HauptState);
            vorsignal_state = eeprom_read_byte(&CV.VorState);
            CheckSave_timerval = timerval;
            CheckSaveState = 1;
            break;
        case 1:
            if (semaphor_get(C_DoSave) )
              {
                if (JUMPER_FITTED)
                  {
                     CheckSave_timerval = timerval;
                     CheckSaveState = 2;
                  }
              }
            break;
        case 2:
            if ((unsigned char)(timerval - CheckSave_timerval) > SAVE_TIMEOUT)
              {
                turn_led_on();
                unsigned char my_mode;
                my_mode = eeprom_read_byte(&CV.MODE);
                switch (my_mode)
                  {
                    default:
                    #if (PORT_ENABLED == TRUE)
                        case 0:
                            RAM_LastState = PortState;  // i.e. PORTB   
                        break;
                    #endif
                        case 32:
                            RAM_LastState = CurrentTarget;
                            break;
                  }
                if (RAM_LastState != eeprom_read_byte(&CV.LastState))
                  {
                    eeprom_write_byte(&CV.LastState, RAM_LastState); 
                  }
                if (hauptsignal_state != eeprom_read_byte(&CV.HauptState))
                  {
                    eeprom_write_byte(&CV.HauptState, hauptsignal_state); 
                  }
                if (vorsignal_state != eeprom_read_byte(&CV.VorState))
                  {
                    eeprom_write_byte(&CV.VorState, vorsignal_state); 
                  }
                CheckSaveState = 1;
                turn_led_off();                
              }
            else
              {
                if (semaphor_get(C_DoSave) )
                  {
                    if (JUMPER_FITTED)
                      {
                         CheckSave_timerval = timerval;         // retrigger timer
                      }
                  }
              }          
            break;
      }
  }



//------------------------------------------------------------------------
// This Routine is called when PROG is pressed
// -- manual programming and accordingly setting of CV's
//
#define DEBOUNCE  (50000L / TICK_PERIOD)
#if (DEBOUNCE == 0)
 #define DEBOUNCE  1
#endif


void DoProgramming(void)
  {
    unsigned char myCommand;
    signed char my_timerval;
    unsigned char pulsdelay;
    
       

    my_timerval = timerval;
    while(timerval - my_timerval < DEBOUNCE) ;          // wait

    if (PROG_PRESSED)                                   // still pressed?
      {
        turn_led_on();
        
        while(PROG_PRESSED) ;                           // wait for release

        my_timerval = timerval;
        while(timerval - my_timerval < DEBOUNCE) ;     // wait
        
        while(!PROG_PRESSED)
          {
            if (semaphor_get(C_Received))
              {                                         // Message
                if (analyze_message(&incoming))                  // yes, any accessory
                  {
                    eeprom_write_byte(&CV.myAddrL, (unsigned char) ReceivedAddr & 0b00111111  );     
                    eeprom_write_byte(&CV.myAddrH, (unsigned char) (ReceivedAddr >> 6) & 0b00000111);
                    
                    myCommand = ReceivedCommand & 0x07;

                    simple_preset(myCommand);

                    // ????
                    
                    if (myCommand == 0)      pulsdelay =  200000L / TICK_PERIOD; 
                    else if (myCommand == 1) pulsdelay =  500000L / TICK_PERIOD; 
                    else if (myCommand == 2) pulsdelay = 1000000L / TICK_PERIOD; 
                    else if (myCommand == 3) pulsdelay = 2000000L / TICK_PERIOD; 
                    else                     pulsdelay = 0;   // permanent; 
               
                    eeprom_write_byte(&CV.T_on_F1, pulsdelay);   
                    eeprom_write_byte(&CV.T_on_F2, pulsdelay);   
                    eeprom_write_byte(&CV.T_on_F3, pulsdelay);   
                    eeprom_write_byte(&CV.T_on_F4, pulsdelay);   
             
                    if (JUMPER_FITTED)
                      {
                        if (pulsdelay == 0)
                          {
                            eeprom_write_byte(&CV.LastState, 0x55);
                            eeprom_write_byte(&CV.FBM_F1, 0);    // 0 = COMMAND_ONLY
                            eeprom_write_byte(&CV.FBM_F2, 0);    // 0 = COMMAND_ONLY
                            eeprom_write_byte(&CV.FBM_F3, 0);    // 0 = COMMAND_ONLY
                            eeprom_write_byte(&CV.FBM_F4, 0);    // 0 = COMMAND_ONLY
                          }
                        else
                          {
                            eeprom_write_byte(&CV.LastState, 0x00);
                            eeprom_write_byte(&CV.FBM_F1, 1);    // 1 = END_SWITCH
                            eeprom_write_byte(&CV.FBM_F2, 1);    // 1 = END_SWITCH
                            eeprom_write_byte(&CV.FBM_F3, 1);    // 1 = END_SWITCH
                            eeprom_write_byte(&CV.FBM_F4, 1);    // 1 = END_SWITCH
                          }
                      }
                    
                    do {} while (!eeprom_is_ready());    // wait for write to complete
                    
                    LED_OFF;

                    // we got reprogrammed ->
                    // forget everthing running and restart decoder!                    
                    
                    cli();
                    
                    // laut diversen Internetseiten sollte folgender Code laufen -
                    // tuts aber nicht, wenn man das Assemblerlistung ansieht.
                    // void (*funcptr)( void ) = 0x0000;    // Set up function pointer
                    // funcptr();                        // Jump to Reset vector 0x0000
                    
                    __asm__ __volatile 
                    (
                       "ldi r30,0"  "\n\t"
                       "ldi r31,0"  "\n\t"
                       "icall" "\n\t"
                     );
                    
                    // return;  
                  }
              }
          }  // while
        turn_led_off();
        my_timerval = timerval;
        while(timerval - my_timerval < DEBOUNCE) ;     // wait    
        while(PROG_PRESSED) ;       // wait for release
      }
    return;   
  }


//--------------------------------------------------------------------------------------------

#define ASM_DUMMY() __asm__ __volatile__ ("" : : )


int main(void)
  {
    unsigned char my_mode;

    init_main();                                        // setup hardware ports (to do!!)

    init_dimm_engine();                                 // setup timers and states

    init_dcc_receiver();                                // setup dcc receiver

    init_dcc_decode();

    init_keyboard();                                    // local tracers

    if (eeprom_read_byte(&CV.myAddrH) & 0x80)
      {
        flash_led_fast(5);                              // warning - we are unprogrammed
      }

    sei();                                              // Global enable interrupts

    CheckSaveState = 0;                                 // force reinit

    while(1)
      {
        if (semaphor_query(C_Received))
          {
            if (analyze_message(&incoming) >= 2)                 // MyAdr or greater received
              {
                my_mode = eeprom_read_byte(&CV.MODE);
                switch (my_mode)
                  {
                    #if (PORT_ENABLED == TRUE)
                        case 0:
                            port_action(ReceivedCommand, ReceivedActivate);   // standard accessory decoder
                        break;
                    #endif
                    
                        case 32:
                            action(ReceivedCommand, ReceivedActivate);
                            break;
                    default:
                        break;
                  }
              }
            semaphor_get(C_Received);                       // now take away the protection
          }

        if (semaphor_get(C_Tick) )
          {
            dimmer();
          } 

        CheckSave();

        /*
        if (semaphor_get(C_DoSave) )
          {
            if (JUMPER_FITTED)
              {
                my_mode = eeprom_read_byte(&CV.MODE);
                switch (my_mode)
                  {
                    #if (PORT_ENABLED == TRUE)
                        case 0:
                            eeprom_write_byte(&CV.LastState, PortState);  // i.e. PORTB   
                        break;
                    #endif
                        case 32:
                            eeprom_write_byte(&CV.LastState, CurrentTarget);  // 
                            break;
                    default:
                        break;
                  }
              }
          } 
        */

        if (PROG_PRESSED) DoProgramming();

      }
  }


//===============================================================================
//
// Simulationen:
//
// 0: keine Simulation
// 1: Test der Empfangrootine; hierzu ein DCC Generator (aus opendcc/dccout.c)
// 2: Test der Portengine
//
#if (SIMULATION == 1)

//                                        Adressbyte   Datenbyte 
//                                         10AAAAAA    1aaaSCCR    // note: aaa = ~AAA
unsigned char message_adr001_out0_g[] = {0b10000001, 0b11111001};
unsigned char message_adr005_out1_g[] = {0b10000101, 0b11111011};
unsigned char message_adr380_out0_g[] = {0b10111100, 0b10101001};  // 380 = 0x17C = 0b101111100


// upstream interface:
struct
  {
    unsigned char size;
    unsigned char dcc[5];
  } next_message;

volatile unsigned char next_message_count;

enum do_states
  {                            // actual state
     dos_idle,
     dos_send_preamble,
     dos_send_bstart,
     dos_send_byte,
     dos_send_xor
  };

struct
  {
    enum do_states state;

    unsigned char ibyte;                            // current index of byte in message
    unsigned char bits_in_state;                    // Bits to output in this state
    unsigned char cur_byte;                         // current byte
    unsigned char xor_byte;                              // actual xor
    unsigned char current_dcc[5];                  // current message in output processing
    unsigned char bytes_in_message;                 // current size of message (decremented)
    unsigned char phase;
  } doi;


unsigned char dccbit;

void do_send(unsigned char mydccbit)
  {
    dccbit = mydccbit;
  }

void dcc_bit_generator(void)
  {
    switch (doi.state)
      {
        case dos_idle:
            do_send(1);
            if (next_message_count > 0)
              {
                memcpy(doi.current_dcc, next_message.dcc, sizeof(doi.current_dcc));
                doi.bytes_in_message = next_message.size;
                // no size checking - if (doi.cur_size > 5) doi.cur_size = 5;
                next_message_count--;
                doi.ibyte = 0;
                doi.xor_byte = 0;
                doi.bits_in_state = 14;
                doi.state = dos_send_preamble;
              }
            break;

        case dos_send_preamble:
            do_send(1);
            doi.bits_in_state--;
            if (doi.bits_in_state == 0)
                 doi.state = dos_send_bstart;
            break;

        case dos_send_bstart:
            do_send(0);
            if (doi.bytes_in_message == 0)
              { // message done, goto xor
                doi.cur_byte = doi.xor_byte;
                doi.state = dos_send_xor;
                doi.bits_in_state = 8;
              }
            else
              { // get next addr or data
                doi.bytes_in_message--;
                doi.cur_byte = doi.current_dcc[doi.ibyte++];
                doi.xor_byte ^= doi.cur_byte;
                doi.state = dos_send_byte;
                doi.bits_in_state = 8;
              }
            break;

        case dos_send_byte:
            if (doi.cur_byte & 0x80) do_send(1);
            else                    do_send(0);
            doi.cur_byte <<= 1;
            doi.bits_in_state--;
            if (doi.bits_in_state == 0)
              {
                doi.state = dos_send_bstart;
              }
            break;

        case dos_send_xor:
            if (doi.cur_byte & 0x80) do_send(1);
            else                    do_send(0);
            doi.cur_byte <<= 1;
            doi.bits_in_state--;
            if (doi.bits_in_state == 0)
              {
                doi.state = dos_idle;
              }
            break;
     }
  }

void dcc_generate_init(void)
  {
    doi.state = dos_idle;
  }


void simulat_receive(void)
  {
    dcc_generate_init();
    dcc_bit_generator();
    dcc_bit_generator();
    memcpy(next_message.dcc, message_adr001_out0_g, sizeof(doi.current_dcc));
    next_message.size = 2;
    next_message_count = 2;
    while(1)
      {
        dcc_bit_generator();
        dcc_receive();

        if (semaphor_get(C_Received) )
          {
            if (analyze_message() == 2)     // MyAdr empfangen
              {
                action();
              }
            semaphor_clear(C_Received);
          }
        if (semaphor_get(C_DoSave) )
          {
            semaphor_clear(C_DoSave);
            if (JUMPER_FITTED)
              {
                eeprom_write_byte(&EE_LastState, PORTB);   
              } 
          } 
        // if (PROG_PRESSED) DoProgramming();         
      }
 
  }


#endif
//------------------------------------------------------------------------
